////////////////////////////////////////////////////////////////////////////////
//
// PackageFormat-Unit:
//
// Includes: - TUTPackage object (main object, used for all other operations)
//
// (c) 2000, by Jesco Topp
////////////////////////////////////////////////////////////////////////////////
unit PackageFormat;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  StdCtrls, ExtCtrls, Forms, Spin, Dialogs, PkgAux, NormalTexture;

////////////////////////////////////////////////////////////////////////////////

//TGroup
type
  TGroup = Class(TObject)
    Name    : string;
    Entries : TStrings;
    constructor Create;
    procedure Free;
    destructor Destroy; override;
end;

////////////////////////////////////////////////////////////////////////////////
//Main-UTPackage-Object

type
  TUTPackage = Class(TObject)
    //Header-Data
    Signature      : DWORD;
    PkgVer         : DWORD;
    PackageFlags   : DWORD;
    //HeaderTables
    NameTable      : TNameTable;
    ImportTable    : TImportTable;
    ExportTable    : TExportTable;
    //List of internal packages
    Group          : array [0..127] of TGroup;
    GroupCount     : integer;
    //Texture related stuff
    NormalTexture  : TNormalTexture; //normal texture (properties, data and palettes)
    TextureCount   : integer;
    //Misc stuff
    PackageSize    : longint;
    Name           : string;
    //Functions
    constructor Create;
    procedure   Free;
    destructor  Destroy; override;
  public
    function    Load(const FileName : string) : Boolean; //Load a package and its headers
    procedure   LoadNormalTexture(const Name : string; const MaxPreviewSize : integer); //Loads a normal texture (meaning no scripted ones like FireTexture)
  private
    F              : File;
    IsPkgOpen      : Boolean;
    procedure CreateGroupList;           //Create a list of internal packages
end;
//////////////////////////////

implementation

////////////////////////////////////////////////////////////////////////////////

//TGroup
constructor TGroup.Create;
begin
  Entries := TStringList.Create;
end;

procedure TGroup.Free;
begin
 if Self <> nil then Destroy;
end;

destructor TGroup.Destroy;
begin
  Entries.Free;

  inherited Destroy;
end;

////////////////////////////////////////////////////////////////////////////////
//TUTPackage

constructor TUTPackage.Create;
var
  i         : integer;
begin
  //Create all objects
  NameTable     := TNameTable.Create;
  ImportTable   := TImportTable.Create;
  ExportTable   := TExportTable.Create;
  NormalTexture := TNormalTexture.Create;
  for i:=0 to 127 do
    Group[i] := TGroup.Create;

  //No file currently opened
  IsPkgOpen := FALSE;
end;

procedure TUTPackage.Free;
begin
 if Self <> nil then Destroy;
end;

destructor TUTPackage.Destroy;
var
  i        : integer;
begin
  //Destroy everything
  NameTable.Free;
  ImportTable.Free;
  ExportTable.Free;
  NormalTexture.Free;
  for i:=0 to 127 do
    Group[i].Free;

  //Check if file is open or not
  if (IsPkgOpen = TRUE) then
    CloseFile(F);
end;

//Loads a package and its headers
function TUTPackage.Load;
var
  i      : integer;
  temp   : DWORD;
begin
  //Setup everything
  AssignFile(F, FileName);
  FileMode := 0;
  Reset(F,1);
  IsPkgOpen := TRUE;
  Name := FileName;

  //Get filesize
  PackageSize := FileSize(F);

  //Read Header and tables
  BlockRead(F, Signature, 4);
  if (Signature <> $9E2A83C1) then
    Result := FALSE
  else
  begin
    BlockRead(F, PkgVer, 4);      //Read PackageVersion
    BlockRead(F, PackageFlags, 4);//Read Package-Flags
    BlockRead(F, NameTable.Count, 4);   //Read NameCount & Offset
    BlockRead(F, NameTable.Offset, 4);
    BlockRead(F, ExportTable.Count, 4); //Read ExportCount & Offset
    BlockRead(F, ExportTable.Offset, 4);
    BlockRead(F, ImportTable.Count, 4); //Read ImportCount & Offset
    BlockRead(F, ImportTable.Offset, 4);

    //Read NameTable
    Seek(F, NameTable.Offset);
    for i := 0 to NameTable.Count - 1 do
    begin
      NameTable.ObjectName.Add(ReadNAME(F, PkgVer));
      BlockRead(F, temp, 4);
      NameTable.ObjectFlags.Add(IntToStr(temp));
    end;

    //Read ImportTable
    Seek(F, ImportTable.Offset);
    for i := 0 to ImportTable.Count - 1 do
    begin
      ImportTable.ClassPackage.Add(NameTable.ObjectName.Strings[ReadINDEX(F)]);
      ImportTable.ClassName.Add(NameTable.ObjectName.Strings[ReadINDEX(F)]);
      BlockRead(F, temp, 4); ImportTable.ImportPackage.Add(IntToStr(temp));
      ImportTable.ObjectName.Add(NameTable.ObjectName.Strings[ReadINDEX(F)]);
    end;

    //Read ExportTable
    Seek(F, ExportTable.Offset);
    for i := 0 to ExportTable.Count - 1 do
    begin
      ExportTable.ExportClass.Add(IntToStr(ReadINDEX(F)));
      ExportTable.Super.Add(IntToStr(ReadINDEX(F)));
      BlockRead(F, temp, 4); ExportTable.ExportPackage.Add(IntToStr(temp));
      ExportTable.ObjectName.Add(NameTable.ObjectName.Strings[ReadINDEX(F)]);
      BlockRead(F, temp, 4); ExportTable.ObjectFlags.Add(IntToStr(temp));
      ExportTable.SerialSize.Add(IntToStr(ReadINDEX(F)));
      if (StrToInt(ExportTable.SerialSize.Strings[i]) > 0) then
        ExportTable.SerialOffset.Add(IntToStr(ReadINDEX(F)))
      else
        ExportTable.SerialOffset.Add('-1');
    end;

    //Get object references (requires complete tables)
    for i := 0 to ImportTable.Count - 1 do
    begin
      ImportTable.ImportPackage.Strings[i] := GetObjectReference(NameTable, ImportTable, ExportTable, StrToInt(ImportTable.ImportPackage.Strings[i]));
    end;
    for i := 0 to ExportTable.Count - 1 do
    begin
      ExportTable.ExportClass.Strings[i] := GetObjectReference(NameTable, ImportTable, ExportTable, StrToInt(ExportTable.ExportClass.Strings[i]));
      ExportTable.Super.Strings[i] := GetObjectReference(NameTable, ImportTable, ExportTable, StrToInt(ExportTable.Super.Strings[i]));
      ExportTable.ExportPackage.Strings[i] := GetObjectReference(NameTable, ImportTable, ExportTable, StrToInt(ExportTable.ExportPackage.Strings[i]));
    end;

    //Create packagelist of internal packages + textures
    CreateGroupList;

    Result := TRUE;
  end;
end;

procedure TUTPackage.LoadNormalTexture;
begin
  //Free previous texture
  NormalTexture.Free;
  NormalTexture := TNormalTexture.Create;

  NormalTexture.Load(NameTable, ImportTable, ExportTable, PkgVer, Name, MaxPreviewSize, F);
end;

//Create a list of internal packages
procedure TUTPackage.CreateGroupList;
var
  i,k,j   : integer;
begin
  k := 0;
  TextureCount := 0;
  //get packages/textures from ExportTable
  for i:=0 to ExportTable.Count-1 do
  begin
    if (ExportTable.ExportClass.Strings[i] = 'Package') then
    begin
      Group[k].Name := ExportTable.ObjectName.Strings[i];
      for j:=0 to ExportTable.Count-1 do
        if (ExportTable.ExportClass.Strings[j] = 'Texture') then
          if  (ExportTable.ExportPackage.Strings[j] = Group[k].Name) then
              Group[k].Entries.Add(ExportTable.ObjectName.Strings[j]);
      k:=k+1;
    end;
  end;

  //Create [ALL] Group
  Group[k].Name := '[ALL]';
  for i:=0 to ExportTable.Count-1 do
  begin
    if (ExportTable.ExportClass.Strings[i] = 'Texture') then
    begin
      Group[k].Entries.Add(ExportTable.ObjectName.Strings[i]);
      TextureCount := TextureCount + 1;
    end;
  end;
  GroupCount := k+1;
end;

////////////////////////////////////////////////////////////////////////////////

end.
